#|BSD License

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. |#



(require :util )

(defpackage :ph_simplex
  (:use :cl :util)
  (:export #:crea_simplex #:min_simplex)
  (:documentation "This package provides the simplex n-variable minimization method.")
)

(in-package :ph_simplex)

(defun ordena_simplex (simplex funcio)
  "Ordena el smplex, deixant a la primera posici el menor element, i aix. Com a arguments se li passen el smplex i una funci lambda ."
  (let ((simp (copy-seq simplex))) ;copie la seqncia per a no cagar-la, ggg
	(reverse (sort simp (lambda (x y) (if (< (apply funcio x) (apply funcio y)) nil t))))))

(defun centre_de_gravetat (simp )
  "Torna el punt centre de gravetat d'un smplex. El smplex ha d'estar ordenat abans de passar-lo a la funci."
  (let* ((primer (nth 0 simp))
	 (dimensio (length primer))
					;(elements (+ dimensio 1)) ; en dimensio 1 el smplex t 2 elements, en dimensi 2, 3 , etc... Per raons d'optimitzaci, use dimensio enlloc i no necessite aquesta varaible.
	 (i 0)
	 (cdg (repeat 0 dimensio)) ;inicialitze el centre de gravetat al vector nul en el nombre de dimensions que usem.
	 )
    (loop
     (if (eq i dimensio) (return) ;L'ltim vector que sumarem s el dimensio - 1 (el que ocupa la posici dimensio s el pitjor)
       (progn 
	 (setq cdg (mapcar #'+ cdg (nth i simp)))
;	 (format t "El centre de gravetat actual s ~A, i li he sumat ~A~%" cdg (nth i simp))
	 (incf i)
	 )))
    (mapcar #'* (repeat (/ 1 i) dimensio ) cdg) ; Ara multiplique cdg per 1/i , per obtindre el resultat final.

     ))



(defun crea_simplex (n_dimensions &rest valors_inicials)
  "Torna un smplex en n_dimensions, tenint en compte valors_inicials."
  (let ((simplex '())
	aux)
    (do ((i 0 (incf i)))
	((= i  n_dimensions ) t)
	(setq aux (repeat 0 n_dimensions))
	(setf (nth i aux) 1)
	(push aux simplex))
    (push valors_inicials simplex)))
    

(defun min_simplex (simplex funcio &key (NITER 100) (tol_rel 1e-3))
  "Minimitza la funci de vries variables funcio per mtode del smplex. Com a arguments rep un smplex a usar i la funci, de forma opcional el nombre d'iteracions i la tolerncia relativa."
  (let* ((smp (copy-seq simplex)) ;Copie el smplex per a evitar modificar-lo. 
	(err 0) ;Variable on posar l'error mxim (en valor absolut) que es tollerar.
	(iteracio 0) ;Variable per guardar el nombre d'iteraci actual.
	(dimensio (length (nth 0 smp)))
	(pos_millor 0)
	(pos_pitjor dimensio)
	(pos_quasipitjor (- dimensio 1))
	millor
	pitjor
	quasipitjor
	cdg
	nou
	nou2
	)

    (setq smp (ordena_simplex smp funcio))
    (setq millor (nth pos_millor smp))
    (setq pitjor (nth pos_pitjor smp))
    ;Assigne l'error com , amb els valors inicials de millor i pitjor, (f(millor) - f(pitjor))*tol_rel
    (setq err (abs (* (- (apply funcio millor) (apply funcio pitjor)) tol_rel)))
;    (format t "Treballant amb un error ~A~%" err)
    (loop
     (if (= iteracio NITER) (return) 
       (progn
	
	 (setq smp (ordena_simplex smp funcio))
	 (setq cdg (centre_de_gravetat smp))
	 (setq millor (nth pos_millor smp))
	 (setq pitjor (nth pos_pitjor smp))
	 (setq quasipitjor (nth pos_quasipitjor smp))
	 (incf iteracio)
					;Una vegada calculades totes les magnituts necessries, em pose a calcular.
					;El punt nou es calcula com :nou = 2*cdg - pitjor;
	 (setq nou (mapcar #'- (mapcar #'* (repeat 2 dimensio) cdg) pitjor))
					;Cas 1:
	 (if (< (apply funcio nou) (apply funcio millor))
	     (progn ;En aquest cas es calcula una altra reflexi
	       (setq nou2 (mapcar #'- (mapcar #'* (repeat 3 dimensio) cdg) pitjor))
;	       (format t "Cas 1~%")
	       (if (< (apply funcio nou2) (apply funcio nou)) ;Cas 1A: En aquest cas tenim un punt molt bo i el canviem el pitjor per nou2
		   (progn
		 ;    (format t "Cas 1A~%")
		     (setf (nth pos_pitjor smp) nou2))
		 (progn ;Cas 1B: En cas contrari posem nou a la posici de pitjor
		;   (format t "Cas 1B~%")
		   (setf (nth pos_pitjor smp) nou))))
	   (progn 
	     (if (< (apply funcio nou) (apply funcio quasipitjor)) ;Cas 2: el nou punt s millor que quasipitjor -> canviem nou per pitjor
		 (progn 
	;	   (format t "Cas 2~%")
		   (setf (nth pos_pitjor smp) nou)))
	     (progn ;Cas 3: El nou punt s pitjor que quasipitjor
;	       (format t "Cas 3~%")
					;Calcule nou2 com nou2 = 0.5*(cdg + pitjor):
	       (setq nou2 (mapcar #'* (repeat 0.5 dimensio) (mapcar #'+ cdg pitjor)))
					;Cas 3a: el nou punt s millor que quasipitjor -> canviem pitjor per nou2
	       (if (< (apply funcio nou2) (apply funcio quasipitjor))
		   (progn 
		 ;    (format t "Cas 3A~%")
		     (setf (nth pos_pitjor smp) nou2))
		 (progn ;Cas 3b: procedim a una contracci general del simplex
		;   (format t "Cas 3B~%")
		   (do ((contador 1 (incf contador)))
		       ((= contador (+ dimensio 1)) t) 
		       (setf (nth contador smp) (mapcar #'* (repeat 0.5 dimensio) (mapcar #'+ millor (nth contador smp))))))))))
	; (format t "Desprs de la iteraci ~D tenim el simplex : ~%" iteracio)
	; (mapcar (lambda (x) (print x)) smp)
	 (setq smp (ordena_simplex smp funcio))
	 (let ((error_pas (abs (- (apply funcio millor) (apply funcio pitjor)))) ;error_pas s l'error que cometem a aquest pas
	       (error_permes (max tol_rel (* tol_rel (/ (+ (apply funcio millor) (apply funcio pitjor)) 2)))))
;	   (format t "L'error en aquest pas s ~A i l'error admissible ~A~%" error_pas error_permes)
	   (if (< (- error_pas error_permes) 0) ;Pense que s millor el nou mtode per calcular l'error, i ams que cal fer la resta, doncs comparar dos nombres molt semblants pot donar problemes
	       (progn
;		 (print "He trobat un mnim")
		 (return-from min_simplex (values t smp (mapcar #'(lambda (x) (if (= 0 x) error_pas (abs x))) (mapcar #'- (nth pos_millor smp) (nth pos_pitjor smp)))))))
	   #|	 (let ((err_tmp (abs (- (apply funcio millor) (apply funcio pitjor)))))
	   (format t "L'error en aquest pas s ~A i l'error admissible ~A~%" err_tmp err)
	   (if (< err_tmp err)
	       (progn 
		 (print "He trobat un mnim")
		 (return-from min_simplex nil))))|#
	   )
	 )))
;    (print "Sortint sense trobar cap mnim")
;    (print smp)
    (values nil smp)))


;NOTA: Cal que torne l'error, com a millor-millor_anterior
(defun drive_simplex (simplex funcio &key (NITER 100) (tol_rel 1e-3) (NMIN 10) (tol_abs 1e-3))
  "Condueix el mtode del smplex, aplicant-lo successivament fins conseguir un resultat acceptable. La idea s la segent:
Es minimitza el smplex. S'agafa el millor punt del nou smplex i es comprova si s millor que millor, en eixe cas el canvie per millor al smplex vell. Si s millor que pitjor, el canvie per pitjor. A continuaci, torne a repetir el procs. I aix fins que f(pitjor) - f(millor) < tol_abs. Torna dos valors: t si s'ha trobat el mnim (nil en cas contrari), el smplex amb el qual ha treballat en l'ltima iteraci realitzada i, en cas de trobar el mnim, una llista amb l'error del mnim, estimat com millor-millor_anterior."
  (let* ((smp (copy-seq simplex))
	(err 0)
	(dimensio (length (nth 0 smp)))
	(pos_millor 0)
	(pos_pitjor dimensio)
;	(pos_quasipitjor (- dimensio 1))
	millor
	pitjor
	;quasipitjor
	(n_minimitzacio 0))
    (loop
     (if (= n_minimitzacio NMIN) (return-from drive_simplex nil)
       (progn
	 (incf n_minimitzacio)
	 ;ULL!! Ac tenim un error!! Cal comprovar si es torna un nil!!!!
	 (multiple-value-bind (trobat_minim smp_tmp errors ) (min_simplex smp funcio :tol_rel tol_rel :NITER NITER) ;Pose a smp_tmp el smplex ordenat resultat de la iteraci
			 ;     (print "S'ha sortint del smplex")
			      (setq smp_tmp (ordena_simplex smp_tmp funcio))
			      (setq millor (nth pos_millor smp_tmp))
			      (setq pitjor (nth pos_pitjor smp_tmp))
			      (setq smp (ordena_simplex smp funcio))
			      (if (< (apply funcio millor) (apply funcio (nth pos_millor smp))) ;Si el nou millor s millor que el vell millor
				  (setf (nth pos_pitjor smp) millor) ;El nou pitjor ser el millor trobat, aix ellimine el punt pitjor.
				(progn ;En cas contrari
				  (if (< (apply funcio millor) (apply funcio (nth pos_pitjor smp))) ;Si el nou millor s millor que el vell pitjor
				      (setf (nth pos_pitjor smp) millor) ;Els canvie
				    (progn ;En cas contrari no he minimitzat ni una mica el smplex, simplement surt sense fer res
				;      (print "Amb tals valors inicials per al smplex no s possible minimitzar-lo. Sortint...")
				      (return-from drive_simplex nil)))))
			      (setq err (abs (-(apply funcio millor) (apply funcio (nth pos_millor smp))))) ;L'error s f(pitjor) - f(millor)
					;Una vegada fets els canvis, mire a veure si puc sortir
			     ; (format t "Comparant error en pas ~A amb error admisible ~A~%" err tol_abs)
			      (if (< (- err tol_abs) 0) ;Si l'error s menor que el tolerat sortim
				  (progn
				;    (format t  "Trobat mnim a Driver (minimitzaci ~D)!!~%" (- n_minimitzacio 1))
				    (return-from drive_simplex (values t (ordena_simplex smp funcio) (mapcar #'(lambda (x) (if (> x tol_abs) x tol_rel)) errors))))
				)
			      )
	 )
       )
     )
;    (print "No s'ha trobat mnim a Driver")
    (values nil (ordena_simplex smp funcio))
    ))
